# register all encoders
import ludwig.encoders.bag_encoders
import ludwig.encoders.category_encoders
import ludwig.encoders.date_encoders
import ludwig.encoders.generic_encoders
import ludwig.encoders.h3_encoders
import ludwig.encoders.image
import ludwig.encoders.sequence_encoders
import ludwig.encoders.set_encoders
import ludwig.encoders.text_encoders  # noqa
