from typing import Any, Dict

# The dictionary representation of a whylogs.core.view.column_profile_view.ColumnProfileView
# Obtained by calling ColumnProfileView.to_summary_dict().
# See profiling/contants for a list of keys.
ColumnProfileSummary = Dict[str, Any]
