# Register all preprocessors
from ludwig.schema.features.preprocessing import (  # noqa
    audio,
    bag,
    binary,
    category,
    date,
    h3,
    image,
    number,
    sequence,
    set,
    text,
    timeseries,
    vector,
)
